
import time
import pyinputplus as pyip

from myClasses import Conversation


def get_conversation_names(record_path):
    """
    Get the names of all the stored conversations.
    If the record file does not exist, it is created, and an empty list is returned.
    Its parent directory is also created if it does not exist.

    Parameters:
    - record_path (Path): A Path object representing the path to the conversation record file.

    Returns:
    - conversation_names (list): A list of strings representing the names of the stored conversations.
    """
    conversation_names = []
    if not record_path.exists(): # If the record file does not exist
        # Create the record file and its parent directory
        print("No existe registro de conversaciones.")
        print("Creando registro...\n")
        record_path.parent.mkdir(parents=True, exist_ok=True)
        record_path.touch(exist_ok=True)
        time.sleep(2)
        
        # Return an empty list
        return conversation_names
    
    else: # If the record file exists
        # Read the record file and get the conversation names
        with record_path.open() as file:
            for name in file:
                conversation_names.append(name.strip())
        
        # Return the list of conversation names
        return conversation_names

def menu(conversation_names):
    """
    Display the menu and get the user's choice in a dictionary.

    Parameters:
    - conversation_names (list): A list of strings representing the names of the stored conversations.

    Returns:
    - output (dict): A dictionary containing the user's choice and other information.
        - new_conv (bool): A boolean indicating whether the user wants to start a new conversation.
        - conv_name (str): A string representing the name of the conversation to continue. None if the user wants to start a new conversation.
    """
    output = {}

    print("\nMENU DE CONVERSACIONES")
    print("----------------------------")
    print("¿Desea iniciar una nueva conversación o continuar una existente?")
    print("\t1. Iniciar una nueva conversación")
    print("\t2. Continuar una conversación existente\n")
    choiceA = pyip.inputInt("Ingrese su elección (1 o 2): ", min=1, max=2)

    if choiceA == 2: # Continue an existing conversation
        if conversation_names: # If there are saved conversations
            new_conv = False
            choiceB = pyip.inputMenu(prompt="Selecciona la conversación que quieres continuar:\n\n",
                                      choices=conversation_names, numbered=True, blank=True)
            conv_name = choiceB
        else: # If there are no saved conversations
            print("No hay conversaciones guardadas.")
            new_conv = True
            conv_name = None
            
    elif choiceA == 1: # Start a new conversation
        new_conv = True
        conv_name = None
    
    # Define the output dictionary
    output['new_conv'] = new_conv
    output['conv_name'] = conv_name
    
    return output

def get_conversation(menu_output, conv_folder):
    """
    Function to start a new or continue an existing conversation
        - If new conversation, prompts the user for the name of the conversation
        - Creates a new Conversation object with the given name
        - Appends the new conversation name to the record file

        - If existing conversation, uses the given name to load the conversation from a file
        - Prints the conversation history if the user wants to review it
    
    Parameters:
    - menu_output (dict): A dictionary containing the user's choice and other information.
        - new_conv (bool): A boolean indicating whether the user wants to start a new conversation.
        - conv_name (str): A string representing the name of the conversation to continue. None if the user wants to start a new conversation.
    - conv_folder (Path): Path to the folder where the conversation files and record file are stored.

    Returns:
    - myconv (Conversation): A Conversation object representing the conversation to start or continue.
    """

    new_conversation = menu_output['new_conv']
    conversation_name = menu_output['conv_name']

    if new_conversation: # Start a new conversation

        # Notify the user that the conversation has started
        print("\nIniciando una nueva conversación...\n")
        time.sleep(2)

        # Prompt the user for the name of the conversation
        # Overwrithe the 'conversation_name' variable that is None
        conversation_name = input("Ingrese el nombre de la nueva conversación: ")
        conversation_name = conversation_name.strip().lower().replace(" ", "_")

        # Create a new conversation object
        myconv = Conversation(conversation_name)

        # Append the conversation name to the record file
        file_path = conv_folder / "conversation_record.txt"
        with file_path.open("a") as file:
            file.write(f"{conversation_name}\n")


    else: # Continue an existing conversation

        # Notify the user that the conversation has started
        print(f"\nContinuando la conversación '{conversation_name}'...\n")
        time.sleep(2)

        # Create conversation object and load the conversation from the file
        file_path = conv_folder / f"{conversation_name}.txt"
        myconv = Conversation(conversation_name)
        myconv.load_conversation(file_path)


       # Ask the user if they want to review the conversation history
        review = input("¿Desea revisar el historial de la conversación? (s/n): ")
        if review.lower() == "s":

            # Use dict_format=True to print the messages as dictionaries
            myconv.print_conversation2(dict_format=False)

            input("Presione enter para continuar...")

    return myconv

def initialize_conversation(project_path, system_prompt):
    """
    Initializes a conversation by:
        - retrieving saved conversation names and printing them
        - prompt the user with a menu to start a new conversation or continue an existing one
        - create a new conversation object or load an existing one
        - return the selected conversation object 

    Args:
        projectPath (str): The path to the project folder.
        options (dict): Dictionary containing options for the conversation (global variables from main.py)

    Returns:
        - conv (Conversaton): The selected conversation object.
    """
    
    conv_folder = project_path / "conversations"
    conv_record = conv_folder / "conversation_record.txt"

    conversation_names = get_conversation_names(conv_record)

    print("\nCONVERSACIONES GUARDADAS:")
    print("----------------------------")
    print(f"{conversation_names}\n")

    menu_output = menu(conversation_names)

    # DELETE THIS BLOCK AFTER TESTING
    # print("\nMenu output:")
    # print(menu_output)
    # print()
    # input("Presione enter para continuar...")
    # DELETE THIS BLOCK AFTER TESTING

    conv = get_conversation(menu_output, conv_folder)

    if menu_output['new_conv']:
        conv.add_message("system", system_prompt)

    return conv

def get_empty_conversation(system_prompt):
    """
    Create an empty conversation object with a random name.
    Add the system prompt to the conversation.

    Returns:
        - conv (Conversation): An empty Conversation object with a random name.
    """
    name = "temp_conversation"
    conv = Conversation(name)
    conv.add_message("system", system_prompt)
    return conv