#! python2
# -*- coding: utf-8 -*-

# Activate the SpeechRecognition module with:
# cd pepperspeechrecognition
# python module_speechrecognition.py --pip 192.168.0.185 (NAO IP)


import naoqi
import os
import sys
import msvcrt
import time
import qi
import json
import subprocess
import paho.mqtt.client as mqtt
import settings

from datetime import datetime


class NaoListener:

    # MQTT Settings
    client_id = "stt_client"
    BROKER_ADDRESS = settings.MQTT_BROKER
    BROKER_PORT = settings.MQTT_PORT

    USERNAME = settings.MQTT_USERNAME
    PASSWORD = settings.MQTT_PASSWORD
    
    topics = settings.MQTT_TOPICS # dict of topics
    
    # NAO Settings
    NAO_IP = settings.NAO_IP
    NAO_PORT = settings.NAO_PORT

    appid = settings.NAO_APP_ID
    behavior = settings.NAO_BEHAVIORS
    nao_focus_bh = appid + "/" + "nao_focus"

    # Listening FLAG
    listening = False

    def __init__(self, session):
        # Session and mqtt client objects
        self.session = session
        self.mqtt_client = mqtt.Client(client_id=self.client_id)
        # if the broker requires authentication, uncomment the following line
        self.mqtt_client.username_pw_set(self.USERNAME, self.PASSWORD)

        # Connect to the NAO robot
        self.session.connect("tcp://" + self.NAO_IP + ":" + str(self.NAO_PORT))

        # Set services to the different modules
        self.mem = self.session.service("ALMemory")
        self.tts = self.session.service("ALTextToSpeech")
        self.ams = self.session.service("ALAnimatedSpeech")
        self.autonomous_life = self.session.service("ALAutonomousLife")
        self.behavior_manager = self.session.service("ALBehaviorManager")
        self.motion = self.session.service("ALMotion")
        self.leds = self.session.service("ALLeds")
        self.SpeechRecognition = self.session.service("SpeechRecognition")
        
        # Subscribe to the SpeechRecognition event and set the callback
        self.event = self.mem.subscriber("SpeechRecognition")
        self.event.signal.connect(self.on_speech_recognition)

        # Set the callbacks for the MQTT client
        self.mqtt_client.on_connect = self.on_connect
        self.mqtt_client.message_callback_add(self.topics['speak'], self.on_speak_message)
        self.mqtt_client.message_callback_add(self.topics['exercise'], self.on_exercise_message)


    def calibrate(self):
        self.SpeechRecognition.start()
        self.SpeechRecognition.setHoldTime(3.0)
        self.SpeechRecognition.setIdleReleaseTime(2.0)
        self.SpeechRecognition.setMaxRecordingDuration(10)
        self.SpeechRecognition.setLookaheadDuration(2.0)
        self.SpeechRecognition.setLanguage("es")
        self.SpeechRecognition.calibrate()
        # self.SpeechRecognition.setAutoDetectionThreshold(5)
    

    def on_speech_recognition(self, value):
        
        # Disable the SpeechRecognition module
        self.SpeechRecognition.disableAutoDetection()
        self.listening = False
        self.set_eyes_color("green")
        

        # Print the recognized speech
        timestamp = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        print("\nTimestamp: {}".format(timestamp))
        print("Speech recognized:")
        # print(type(value))
        print(value.decode('utf-8'))

        # Create the payload dictionary
        payload = {
            'role': 'user',
            'content': value.rstrip()
        }

        # Convert the payload to JSON format
        payload_json = json.dumps(payload)

        # Publish the message to the 'speech' topic
        self.mqtt_client.publish(self.topics['speech'], payload_json)


    def on_exercise_message(self, client, userdata, message):
        """ Message callback:
        """
        # Disable the SpeechRecognition module
        self.SpeechRecognition.disableAutoDetection()
        self.listening = False
        self.set_eyes_color("green")

        # Get the topic and the message data (decoded from JSON format)
        topic = message.topic
        data = json.loads(message.payload)

        # Get the necessary variables
        timestamp = datetime.now().strftime('%Y-%m-%d %H:%M:%S')
        reps = data['reps'] if data['reps'] else 4
        behavior_str = self.appid + "/" + self.behavior[data['exercise']]
        
        print("\nTimestamp: {}".format(timestamp))
        print("Executing behavior: {}".format(behavior_str))
        print("Number of repetitions: {}".format(reps))


        self.ams.say("Vamos a hacer ejercicios, dame un segundo para prepararme.")
        time.sleep(2)
        # Insert in memory the number of repetitions
        self.mem.insertData("MyApplication/reps", reps)
        # Deactivate the autonomous life
        self.autonomous_life.setAutonomousAbilityEnabled("All", False)
        self.autonomous_life.setState("disabled")
        self.motion.wakeUp()
        time.sleep(1)
        # Start the behavior
        self.behavior_manager.runBehavior(behavior_str)

        # Wait for the behavior to finish
        time.sleep(3)
        print("Behaviour finished")

        # HACER UNA FUNCION DE EL CODIGO SIGUIENTE LLAMADA - NAO IDLE
        # Reactivate the autonomous life
        print("Re-enabling autonomous life...")
        self.autonomous_life.setState("solitary")
        self.autonomous_life.setAutonomousAbilityEnabled("All", True)
        self.autonomous_life.setAutonomousAbilityEnabled("AutonomousBlinking", False)

        # Switch focus back to the default focus
        self.autonomous_life.switchFocus(self.nao_focus_bh)
        time.sleep(2)

        # Reactivate the SpeechRecognition module
        self.SpeechRecognition.enableAutoDetection()
        self.listening = True
        self.set_eyes_color("blue")


    def on_connect(self, client, userdata, flags, rc):
        """ Connection callback, publishes a connection log message """
        if rc == 0:
            # Successful connection & log message
            print("Connected successfully to MQTT broker")
            conn_msg = "Client '{}' connected to MQTT broker".format(self.client_id)
            client.publish(self.topics['log'], conn_msg)

            # Client subscriptions
            client.subscribe(self.topics['speak'])
            client.subscribe(self.topics['exercise'])
        else:
            print("Falied to connect to MQTT broker")

    def on_speak_message(self, client, userdata, message):

        # Extract message information
        topic = message.topic
        data = json.loads(message.payload)
        timestamp = datetime.now().strftime("%Y-%m-%d %H:%M:%S")

        # Extract role and content from the message dictionary
        role = data['role']
        content = data['content']

        # Print message information to the terminal
        print("\nTimestamp: {}".format(timestamp))
        print(u"From Topic: {}".format(topic))
        print(u"{}: {}".format(role, content))

        # Speak the received message
        #self.tts.say(content.encode('utf-8')) # NAOqi requires utf-8 encoding
        self.ams.say(content.encode('utf-8')) # NAOqi requires utf-8 encoding
        time.sleep(1)

        # Reactivate the SpeechRecognition module
        self.SpeechRecognition.enableAutoDetection()
        self.listening = True
        self.set_eyes_color("blue")


    def disconnect_with_msg(self):
        """ Disconnect from the MQTT broker and publishes a disconnection log message """
        # Publish a disconnection log message
        print("\nDisconnecting from the MQTT broker...")
        disconn_msg = "Client '{}' disconnected from MQTT broker".format(self.client_id)
        self.mqtt_client.publish(self.topics['log'], disconn_msg)
        
        # Disconnect from the MQTT broker and stop the loop
        self.mqtt_client.disconnect()
        self.mqtt_client.loop_stop()

    def init_nao(self):
        """ Initialize the NAO robot 
            - Set the language of the TextToSpeech module
            - Deactivate the autonomouslife
        """
        self.tts.setLanguage("Spanish")
        self.motion.wakeUp()
        time.sleep(1)

        self.autonomous_life.setState("solitary")
        self.autonomous_life.stopAll()
        self.autonomous_life.setAutonomousAbilityEnabled("All", True)
        self.autonomous_life.setAutonomousAbilityEnabled("AutonomousBlinking", False)
        self.set_eyes_color("red") 
        self.autonomous_life.switchFocus(self.nao_focus_bh)
        
        print("NAO robot initialized")

    
    def set_eyes_color(self, color):
        self.leds.fadeRGB("FaceLeds", color, 0.3)
    
    def turn_of_robot(self):
        self.SpeechRecognition.disableAutoDetection()
        self.listening = False
        self.set_eyes_color("white")
        self.autonomous_life.setAutonomousAbilityEnabled("All", True)
        self.autonomous_life.stopAll()
        # self.motion.rest()
        self.autonomous_life.setState("disabled")

    def start(self):
        # Connect to the MQTT broker
        print("Connecting to the MQTT broker...")
        self.mqtt_client.connect(self.BROKER_ADDRESS, self.BROKER_PORT)
        # Start the MQTT client loop
        self.mqtt_client.loop_start()
        time.sleep(3)

        self.listening = True
        self.SpeechRecognition.enableAutoDetection()
        self.set_eyes_color("blue")

        try:
            while True:
                time.sleep(1)
                key = get_key()
                # if q is pressed, raise KeyboardInterrupt to exit the program
                if key and key.lower() == b'q':
                    raise KeyboardInterrupt
                # if m is pressed, toggle the listening status of the robot
                elif key and key.lower() == b'm':
                    if self.listening:
                        self.SpeechRecognition.disableAutoDetection()
                        self.set_eyes_color("red")
                        self.listening = False
                    else:
                        self.SpeechRecognition.enableAutoDetection()
                        self.set_eyes_color("blue")
                        self.listening = True
        
        except KeyboardInterrupt:
            print("\nInterrupted by user, shutting down...")
            self.turn_of_robot()
            self.disconnect_with_msg()
            self.session.close()
            print("Session closed")
            print("\nProgram terminated by user...")
            sys.exit(0)


def run_sr_module(folder_path, script_name, options):
    command = 'cd /D {0} && python {1} {2}'.format(folder_path, script_name, options)
    # print("\nRunning command: " + command)
    subprocess.Popen(['cmd', '/C', 'start', 'cmd', '/K', command])


def initiate_sr_module():
    # Get necessary parameters
    nao_ip = settings.NAO_IP
    current_path = os.path.dirname(os.path.realpath(__file__))
    script_name = "module_speechrecognition.py"
    options = "--pip " + nao_ip
    folder_path = os.path.join(current_path, "pepperspeechrecognition")

    # Run the SpeechRecognition module in a new terminal
    process = run_sr_module(folder_path, script_name, options)

    # Return the process
    return process

def get_key():
    """
    Get a key press from the user. Non-blocking.

    Returns:
        str: The key pressed by the user.
    """
    if sys.platform == 'win32':
        if msvcrt.kbhit():
            key = msvcrt.getch()
            return key


def main():

    # Start the SpeechRecognition module
    process = initiate_sr_module()
    
    # Wait for the SpeechRecognition module to start
    print("\nWaiting for the SpeechRecognition module to start...")
    time.sleep(5)

    # Create a qi session
    session = qi.Session()

    # Instantiate the NaoListener object and calibrate the SpeechRecognition module
    listener = NaoListener(session)
    listener.calibrate()
    print("\nCalibrating the SpeechRecognition module...")
    time.sleep(6)

    # Initialize the NAO robot
    listener.init_nao()

    # Start the listener loop
    raw_input("\nPress Enter to start the listener...")
    listener.start()

if __name__ == "__main__":
    main()